#!/usr/bin/perl -w

# Copyright 2016, 2017, 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use warnings;
use Carp 'croak';
use FindBin;
use FLAT;
use List::Util 'sum';
use Math::BaseCnv 'cnv';
use Math::PlanePath::AlternatePaper;
use Memoize;
use Test::More;

use lib $FindBin::Bin;
use MyFLAT ':all';
use MyAlternate ':all';

plan tests => 573;

# uncomment this to run the ### lines
# use Smart::Comments;

# check that the expand(0) and expand(1) of $conf have all of $conf as subsets
sub ok_conf_expand_within_self {
  my ($conf) = @_;
  foreach my $bit (0,1) {
    ok($conf->is_subset( $conf->expand($bit) ),
       "$conf->{'name'}, bit=$bit");
  }
}


#------------------------------------------------------------------------------
# A073057 concats of GRS4

{
  is ($#A073057_HtoL_state_table, 6);
  is ($#A073057_LtoH_state_table, 30);

  # minimum 2 symbols
  my $minimum = FLAT::Regex->new('(0|1) (0|1) (0|1)*')->as_dfa;

  foreach my $want (1 .. 4) {
    my $HtoL = A073057_HtoL_flat($want);
    my $LtoH = A073057_LtoH_flat($want);
    ok (! $HtoL->is_empty);
    ok (! $LtoH->is_empty);

    my $h = $HtoL->intersect($minimum)
      ->MyFLAT::set_name("HtoL $want min 2");
    my $l = $LtoH->MyFLAT::reverse
      ->MyFLAT::set_name("LtoH rev $want");
    ok ($h->equals($l));
    # MyFLAT::FLAT_check_is_equal($h, $l);
  }
}

sub A073057_LtoH_vertical_flip_name {
  my ($name) = @_;
  if (defined $name) {
    if ($name =~ /[a-f]/) {
      $name =~ tr/1234/2143/;
      return $name;
    }
  }
  return undef;
}

my @A073057_LtoH_vertical_flip;
foreach my $state (1 .. $#A073057_LtoH_state_table) {
  my $name = $A073057_LtoH_state_to_name[$state];
  $name = A073057_LtoH_vertical_flip_name($name);
  if (defined $name) {
    $A073057_LtoH_vertical_flip[$state] = $A073057_LtoH_name_to_state{$name};
  }
}

{
  # LtoH same structure in 1<->2 and 3<->4 flip
  my $count = 0;
  foreach my $s (1 .. $#A073057_LtoH_state_table) {
    my $t = $A073057_LtoH_vertical_flip[$s] // next;
    is ($A073057_LtoH_vertical_flip[$t], $s,
        'A073057_LtoH_vertical_flip twice back to original');
    foreach my $bit (0,1) {
      my $s_to = $A073057_LtoH_state_table[$s]->[$bit];
      my $t_to = $A073057_LtoH_state_table[$t]->[$bit];
      is($t_to, $A073057_LtoH_vertical_flip[$s_to],
        "A073057_LtoH_vertical_flip"
         . " s=$A073057_LtoH_state_to_name[$s]\[$s]"
         . " t=$A073057_LtoH_state_to_name[$t]\[$t]"
         . " bit=$bit"
         . "  s_to=$A073057_LtoH_state_to_name[$s_to]\[$s_to]"
         . "  t_to=$A073057_LtoH_state_to_name[$t_to]\[$t_to]");
    }
    $count++;
  }
  is ($count, 24, "A073057_LtoH_vertical_flip pairings");
  is (26, 4*6 + 2);
}


#------------------------------------------------------------------------------
# Palindromes in GRS and More
#
# Jean-Paul Allouche, "Schrodinger Operators With Rudin-Shapiro Potentials
# Are Not Palindromic"

# Return true if $n is the middle of a palindrome of length $len in $func->().
#   even: X  Y  Y  X     odd:  X  Y  Z  Y  X
#           $n                      $n
sub is_func_palindrome_middle {
  my ($n, $len, $func, $start) = @_;
  if ($n < $start) { return 0; }
  $len >= 0 or die;
  my $pre = $n - ($len & 1 ? 1 : 0);  # 0 even, -1 odd
  my $post = $n + 1;                  # +1 always
  for (;;) {
    if ($len < 2) { return 1; }
    $len -= 2;
    if ($pre < $start) { return 0; }
    if ($func->($pre) != $func->($post)) { return 0; }
    $pre--;
    $post++;
  }
  return 1;
}

# Return a FLAT::DFA matching bit string n where TurnLpred(n) == $value.
sub TurnLpred_flat_value {
  my ($value) = @_;
  if ($value==1) { return TurnLpred_flat(); }
  if ($value==0) { return TurnRpred_flat(); }
  die;
}

# Return a FLAT::DFA matching bit string n where TurnRpred(n) == $value.
sub TurnRpred_flat_value {
  my ($value) = @_;
  if ($value==1) { return TurnRpred_flat(); }
  if ($value==0) { return TurnLpred_flat(); }
  die;
}

{
  # Jean-Paul Allouche, "Schrodinger Operators With Rudin-Shapiro Potentials
  # Are Not Palindromic", remark at end of section IV.

  # Dean
  # O E O E odd E O E O   length 9
  # turns at even palindrome length 4

  my $any  = FLAT::Regex->new('(0|1)*')->as_dfa;
  my $zero = MyFLAT::zero_digits_flat(2);

  foreach my $elem
    (
     # TurnLpred = 1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 0, 0, 1, 1, 0, 1, 1, 0
     #             1  2  3  4  5  6
     { name         => 'TurnLpred',
       func         => \&TurnLpred,
       flat_func    => \&TurnLpred_flat_value,
       start        => 1,
       values       => [0,1],
       want_lengths => [1..7, 9, 11, 13],   # evens to 6, odds to 13
       unchanged_lengths => [5,9,13],
     },
     { name         => 'TurnRpred',
       func         => \&TurnRpred,
       flat_func    => \&TurnRpred_flat_value,
       start        => 1,
       values       => [0,1],
       want_lengths => [1..7, 9, 11, 13],
       unchanged_lengths => [5,9,13],
     },

     # dir4 = 0,1,0,3, 0,1,2,1
     #        0 1 2 3  4
     { name         => 'dir4',
       func         => \&dir4,
       flat_func    => \&dir4_flat,
       values       => [0..3],
       want_lengths => [1, 3, 5, 7],
     },
     { name         => 'GRS4',
       func         => \&GRS4,
       flat_func    => \&GRS4_flat,
       values       => [0..3],
       want_lengths => [1, 3, 5, 7],
     },

     { name         => 'Dean alpha',
       func         => \&Dean,       # index from 0
       flat_func    => \&Dean_flat,
       start        => 1,
       values       => [1..4],
       want_lengths => [1, 3, 5, 7, 9, 11, 13],
       unchanged_lengths => [ 5,    9,     13],
     },

     { name         => 'GRS',
       func         => \&GRS,
       flat_func    => \&GRS_flat,
       values       => [1,-1],
       want_lengths => [1..8, 10, 12, 14],
       unchanged_lengths => [6,10,14],
     },
     { name         => 'GRSalt',
       func         => \&GRSalt,
       flat_func    => \&GRSalt_flat,
       values       => [1,-1],
       want_lengths => [1..8, 10, 12, 14],
       unchanged_lengths => [6,10,14],
     },
    ) {
    my $name         = $elem->{'name'};
    my $values       = $elem->{'values'};
    my $func         = $elem->{'func'};
    my $start        = $elem->{'start'} || 0;
    my $want_lengths = $elem->{'want_lengths'};
    my %unchanged_lengths = map {$_=>1} @{$elem->{'unchanged_lengths'}};
    ### $name

    my @len_has_palindromes;
    foreach my $len (@$want_lengths) {
      $len_has_palindromes[$len] = 1;
    }

    my $bad = 0;
  OUTER: foreach my $parity (1,0) {
      # $f = FLAT::DFA matching those n which are the middle of a palindrome
      # of length $len
      #      X  Y  Y  X       X  Y  Z  Y  X
      #         $f                 $f
      # Initially any n for length 0 (even) or 1 (odd).
      #
      my $palindromes = $any;
      if ($start == 1) { $palindromes = $palindromes->difference($zero); }
      my $len = $parity;
      my $pre_offset = ($parity ? 1 : 0);  # 1 odd, 0 even
      my $post_offset = 1;
      my $prev_palindromes;

      while ($len < 20) {
        {
          # those n with the same value at $pre_offset and $post_offset,
          # being union of those with a particular $value at both places
          my $u;
          foreach my $value (@$values) {
            ### $value
            ### $pre_offset
            ### $post_offset
            my $g = $elem->{'flat_func'}->($value);
            my $pre = $g->MyFLAT::digits_increment(add => $pre_offset)
              ->MyFLAT::optional_leading_0s
              ->MyFLAT::minimize;
            my $post = $g->MyFLAT::digits_increment(add => -$post_offset)
              ->MyFLAT::optional_leading_0s
              ->MyFLAT::minimize;
            my $int = $pre->intersect($post)->MyFLAT::minimize;
            ### int num states: $int->num_states
            if ($u) {
              $u = $u->union($int)->MyFLAT::minimize;
            } else {
              $u = $int;
            }
          }
          $palindromes = $palindromes->intersect($u)->MyFLAT::minimize;
          ### u num states: $u->num_states
          ### palindromes num states: $palindromes->num_states
          # $u->MyFLAT::view;
        }

        $len += 2;
        $pre_offset++;
        $post_offset++;

        diag "$name palindromes len=$len num states ", $palindromes->num_states,
          "  empty=", $palindromes->is_empty ? 1 : 0;
        is ($palindromes->is_empty ? 0 : 1,
            $len_has_palindromes[$len] ? 1 : 0,
            "$name has palindromes, len=$len");

        foreach my $n (0 .. 1024) {
          my $n2 = cnv($n,10,2);
          my $by_flat = $palindromes->contains($n2) ? 1 : 0;
          my $by_func = is_func_palindrome_middle($n,$len,$func,$start) ?1:0;
          unless ($by_flat == $by_func) {
            # $palindromes->MyFLAT::view;
            diag "$name len=$len n=$n [$n2] by_flat $by_flat by_func $by_func\n";
            exit;
            last OUTER if $bad++ > 10;
          }
          #      if ($by_flat) { print "$n,"; }
        }

        if ($prev_palindromes) {
          ok ($palindromes->is_subset_of($prev_palindromes),
              "$name length=$len palindromes subset of previous length");
          is ($palindromes->equals($prev_palindromes) ? 1 : 0,
              $unchanged_lengths{$len} ? 1 : 0,
              "$name length=$len palindromes same or diff previous length");
        }

        last if $palindromes->is_empty;

        # if ($len == 14) {
        #   $f->MyFLAT::view;
        #   my $r = $f->MyFLAT::reverse->MyFLAT::minimize;
        #   $r->MyFLAT::view;
        # }
        # exit;

        $prev_palindromes = $palindromes;
      }
    }
    is ($bad, 0, "$name palindromes");
  }
}
# GRS repeats
# select(n->GRS(n)==GRS(n+1), [0..32])
# 0,1,4,7,8,9,11,12,16,17,20,24,25,27,28,31,32
# not in OEIS: 0,1,4,7,8,9,11,12,16,17,20,24,25,27,28,31,32
# GRS(0) == 1
# GRS(1) == 1
# GRS(2) == 1
# GRS(3) == -1


#------------------------------------------------------------------------------
# Dean_flat()

{
  my @want = (1,2,3,4,1,4,3,2,1,2,3,2,1,4,3,4,1,2,3,4,1,4,3,4,1,2,3,2,1,4,3,2,
              1,2,3,4,1,4,3,2,1,2,3,2,1,4,3,2);
  foreach my $i (0 .. $#want) {
    my $n = $i + 1;
    is (Dean($n), $want[$i], 'Dean() vs sample values');
  }
}
{
  my $bad = 0;
  foreach my $n (1 .. 256) {
    my $alpha  = Dean($n);
    my $str = sprintf '%b', $n;
    foreach (0 .. 2) {  # leading zeros
      foreach my $value (1 .. 4) {
        my $by_flat = (Dean_flat($value)->contains($str) ? 1 : 0);
        my $by_func = ($alpha == $value ? 1 : 0);
        unless ($by_flat == $by_func)  { $bad++; }
      }
      $str = "0$str";
    }
  }
  is($bad, 0, "Dean_flat() vs Dean()");
}

foreach my $value (1 .. 4) {
  foreach my $offset (0 .. 3) {
    my $HtoL = Dean_flat($value,$offset);
    # $HtoL->MyFLAT::view;
    my $LtoH = Dean_LtoH_flat($value,$offset);
    # $LtoH->MyFLAT::view;
    ok($HtoL->equals($LtoH->MyFLAT::reverse),
       "Dean_LtoH_flat() vs Dean_flat() value $value offset $offset");
  }
}


#------------------------------------------------------------------------------
# dir4_base4_flat()

{
  my $bad = 0;
  foreach my $d (0 .. 3) {
    my $f = dir4_base4_flat($d);
  N: foreach my $n (0 .. 256) {
      my $str = cnv($n,10,4);
      $str =~ s/^0+//g;
      foreach my $zeros (0,1,3) {
        my $str = ('0'x$zeros) . $str;
        {
          my $by_flat = $f->contains($str) ? 1 : 0;
          my $by_dir  = (dir4($n) == $d);
          unless ($by_flat == $by_dir) {
            print "N=$n  $f->{name} on str=$str got=$by_flat want=$by_dir\n";
            last N if $bad++ > 10;
          }
        }
      }
    }
  }
  is($bad, 0, "dir4_base4_flat() vs dir4()");
}


#------------------------------------------------------------------------------
# GRS4_flat()

{
  my $bad = 0;
  foreach my $d (0 .. 3) {
    my $f = GRS4_flat($d);
  N: foreach my $n (0 .. 256) {
      my $str = sprintf '%b', $n;
      $str =~ s/^0+//g;
      foreach my $zeros (0,1,3) {
        my $str = ('0'x$zeros) . $str;
        {
          my $by_flat = $f->contains($str) ? 1 : 0;
          my $by_GRS  = (GRS4($n) == $d);
          unless ($by_flat == $by_GRS) {
            print "N=$n  $f->{name} on str=$str got=$by_flat want=$by_GRS\n";
            last N if $bad++ > 10;
          }
        }
      }
    }
  }
  is($bad, 0, "GRS4_flat() vs GRS4()");
}


#------------------------------------------------------------------------------
# dir n to n+9 requiring a full set 10 to see all directions mod 4

sub dir4_flat_next {
  my ($diff,$value) = @_;
  return dir4_flat($value)
    ->MyFLAT::digits_increment(add => -$diff)
    ->MyFLAT::optional_leading_0s
    ->MyFLAT::set_name("dir $value at n+$diff")
    ->MyFLAT::minimize (verbose=>1);
}
BEGIN { memoize('dir4_flat_next', LIST_CACHE=>'MERGE') }

# Return true if dir4($n+9) does not occur in dir4($n .. $n+8)
sub dir_n_is_fullset_10 {
  my ($n) = @_;
  my $d = dir4($n+9);
  foreach my $i ($n .. $n+8) {
    if (dir4($i) == $d) { return 0; }
  }
  return 1;
}

{
  # fullset 10
  #     . . . . .  . . . . d
  # takes until n+9 to see final value,
  # so 9 of not d and final d

  my $rpred = TurnLpred_flat()
    ->MyFLAT::digits_increment(add => -1)
    ->MyFLAT::optional_leading_0s
    ->MyFLAT::concat(FLAT::Regex->new('1101'));  # 13 mod 16
  my $lpred = TurnRpred_flat()
    ->MyFLAT::digits_increment(add => -1)
    ->MyFLAT::optional_leading_0s
    ->MyFLAT::concat(FLAT::Regex->new('1010'));  # 10 mod 16
  my $preds = $lpred->union($rpred)
    ->MyFLAT::minimize(verbose=>1);;
  my $preds_LtoH = $preds->MyFLAT::reverse
    ->MyFLAT::minimize(verbose=>1);
  # $preds_LtoH->MyFLAT::view;

  my $u;
  foreach my $value (0 .. 3) {
    ### $value
    my $f = dir4_flat_next(9,$value);
    foreach my $i (0 .. 8) {
      my $sub = dir4_flat_next($i,$value);
      $f = $f->difference($sub)
        ->MyFLAT::set_name("not at $i")
        ->MyFLAT::minimize(verbose=>1);
    }
    if ($u) {
      $u = $u->union($f)
        ->MyFLAT::set_name("need 10 for some value to $value")
        ->MyFLAT::minimize(verbose=>1);
    } else {
      $u = $f;
    }
  }
  $u->MyFLAT::set_name("dir need 10 for all values");
  my $u_LtoH = $u->MyFLAT::reverse
    ->MyFLAT::set_name("dir need 10 for all values, LtoH")
    ->MyFLAT::minimize;

  # $u_LtoH->MyFLAT::view;
  if (0) {
    foreach my $n (0 .. 128) {
      my $n2 = cnv($n,10,2);
      if ($u->contains($n2)) { print "$n,"; }
    }
    print "\n";
  }

  my $bad = 0;
 N: foreach my $n (0 .. 1024) {
    my $by_func = dir_n_is_fullset_10($n) ? 1 : 0;
    my $str = sprintf '%b', $n;
    $str =~ s/^0+//g;
    foreach my $zeros (0,1,3) {
      my $str = ('0'x$zeros) . $str;
      my $by_flat = $u->contains($str) ? 1 : 0;
      unless ($by_flat == $by_func) {
        print "fullset10 n=$n [$str] by flat $by_flat by_func $by_func\n";
        print "  ";
        foreach my $i ($n .. $n+9) {
          print dir4($i),", ";
        }
        print "\n";
        last N if $bad++ > 10;
      }
    }
  }
  is($bad, 0, "flat vs dir_n_is_fullset_10()");

  ok ($preds->equals($u));
  ok ($preds_LtoH->equals($u_LtoH));
  # MyFLAT::FLAT_check_is_equal($preds_LtoH,$u_LtoH);
  # MyFLAT::FLAT_check_is_equal($preds,$u);
}


# # return a FLAT::DFA which is those n with dir4(n)=$value and the next
# # occurrence of $value is at dir4(n+$diff) = $value.
# sub dir_n_difference_flat {
#   my ($diff, $value) = @_;
#   $diff >= 0 or die;
#   my $f     = dir4_flat($value);
#   my $ret = $f;
#   foreach my $i (1 .. $diff-1) {    # n+1 .. n+diff-1
#     my $inc = dir4_flat_next($i,$value);
#     $ret = $ret->difference($inc)  # subtract those with $value at n + $i
#       ->MyFLAT::minimize;
#   }
#   my $inc = dir4_flat_next($diff,$value);
#   return $ret->intersect($inc)
#     ->MyFLAT::set_name("dir $value and diff $diff")
#     ->MyFLAT::minimize (verbose=>1);
# }
# BEGIN { memoize('dir_n_difference_flat', LIST_CACHE=>'MERGE') }
# 
# # Return true if dir($n)==$value and next $value only at dir($n+$diff)==$value.
# sub dir_n_is_difference {
#   my ($n, $diff,$value) = @_;
#   dir($n)%4 == $value or return 0;
#   foreach my $o ($n+1 .. $n+$diff-1) {
#     if (dir($o)%4 == $value) { return 0; }
#   }
#   return dir($n+$diff)%4 == $value;
# }
# {
#   foreach my $diff (0 .. 12) {
#     my $bad = 0;
#     foreach my $value (0 .. 3) {
#       my $f = dir_n_difference_flat($diff,$value);
#       if ($diff % 2) {
#         ok ($f->is_empty,
#             "dir() no repeat at odd difference");
#       }
#     N: foreach my $n (0 .. 256) {
#         my $by_func = dir_n_is_difference($n,$diff,$value) ? 1 : 0;
#         my $str = sprintf '%b', $n;
#         $str =~ s/^0+//g;
#         foreach my $zeros (0,1,3) {
#           my $str = ('0'x$zeros) . $str;
#           my $by_flat = $f->contains($str) ? 1 : 0;
#           unless ($by_flat == $by_func) {
#             print "diff=$diff value=$value n=$n [$str] by flat $by_flat by_func $by_func\n";
#             print "  ";
#             foreach my $i ($n .. $n+$diff) {
#               print dir($i)%4,", ";
#             }
#             print "\n";
#             last N if $bad++ > 10;
#           }
#         }
#       }
#     }
#     is($bad, 0, "dir_n_difference_flat() vs dir_n_is_difference()  diff=$diff");
#   }
# }

# This bit wrong:
# {
#   my $first;
#   foreach my $value (0 .. 3) {
#     ### $value
#     my $f = dir_n_difference_flat(10,$value);
#     if ($first) {
#       $first = $first->union($f)
#         ->MyFLAT::set_name("first of diff 10 values to $value")
#         ->MyFLAT::minimize(verbose=>1);
#     } else {
#       $first = $f;
#     }
#   }
#   $first->MyFLAT::set_name("first of diff 10 any value");
# 
#   foreach my $n (0 .. 128) {
#     my $n2 = cnv($n,10,2);
#     if ($first->contains($n2)) { print "$n,"; }
#   }
#   print "\n";
# 
#   my $last = $first
#     ->MyFLAT::digits_increment(add => 1)
#     ->MyFLAT::optional_leading_0s
#     ->MyFLAT::set_name("n+9 is last of diff 10 any value")
#     ->MyFLAT::minimize(verbose=>1);
# 
#   my $u = $first->intersect($last)
#     ->MyFLAT::set_name("need 10 consecutive")
#     ->MyFLAT::minimize(verbose=>1);
#   $u->MyFLAT::view;
# }



  
  


#------------------------------------------------------------------------------
# GRS

# Return a FLAT::DFA which is those n with GRS(n)=$value and the next
# occurrence of $value is at GRS(n+$diff) = $value.
sub GRS_n_difference_flat {
  my ($diff, $value) = @_;
  my $f   = GRS_flat($value);
  my $opp = GRS_flat(-$value);
  my $ret = $f;  # n
  foreach my $add (1 .. $diff-1) {    # n+1 .. n+diff-1
    $ret = $ret->intersect($opp->MyFLAT::digits_increment(add=>$add));
  }
  return $ret->intersect($f->MyFLAT::digits_increment(add=>$diff)); # n+diff
}
{
  # GRS intersection n to n+4 inclusive is empty.
  # n to n+3 or smaller is infinite.
  # So GRS has runs of at most 4 consecutive values.
  foreach my $value (1, -1) {
    foreach my $diff (1 .. 7) {
      my $f = GRS_n_difference_flat($diff,$value);
      is (!!$f->is_infinite, $diff <= 5, "GRS diff=$diff value=$value");
    }
  }
}

{
  # GRS intersection n to n+4 inclusive is empty.
  # n to n+3 or smaller is infinite.
  # So GRS has runs of at most 4 consecutive values.
  foreach my $value (1,-1) {
    my $f = GRS_flat($value);
    my $int = $f;
    foreach my $add (1,2,3,4) {
      $int = $int->intersect($f->MyFLAT::digits_increment(add=>$add));
      # $int = $int->MyFLAT::minimize;
      # $int->MyFLAT::view;
      is (!!$int->is_infinite, $add <= 3, "$f->{name}  add=$add");
    }
  }
}

{
  my $bad = 0;
 N: foreach my $n (0 .. 256) {
    my $grs = GRS($n);
    my $str = sprintf '%b', $n;
    $str =~ s/^0+//g;
    my $fplus = GRS_flat(1);
    my $fminus = GRS_flat(-1);

    foreach my $zeros (0,1,3) {
      my $str = ('0'x$zeros) . $str;
      my $is_fplus = $fplus->contains($str);
      my $is_fminus = $fminus->contains($str);
      if ($is_fplus == $is_fminus) {
        print "N=$n  GRS_flat +1,-1 are equal\n";
        last N if $bad++ > 10;
      }
      my $grs_by_flat = $is_fplus ? 1 : -1;
      unless ($grs_by_flat == $grs) {
        print "N=$n [$str] GRS=$grs by flat GRS=$grs_by_flat\n";
        last N if $bad++ > 10;
      }
    }
  }
  is($bad, 0, "GRS_flat() vs GRS()");
}


#------------------------------------------------------------------------------
# TurnRunSpred_LtoH_flat()

{
  my $f_one = FLAT::Regex->new ('0* 1')->as_dfa;
  my $f_0m4 = FLAT::Regex->new ('(0|1)* 00')->as_dfa;
  my $f_1m4 = FLAT::Regex->new ('(0|1)* 01')->as_dfa;
  my $f_2m4 = FLAT::Regex->new ('(0|1)* 10')->as_dfa;
  my $f_3m4 = FLAT::Regex->new ('(0|1)* 11')->as_dfa;

  my $case_0m4 = TurnLpred_flat()->intersect($f_0m4);
  my $case_2m4 = TurnRpred_flat()->intersect($f_2m4);
  my $case_1m4 = TurnRpred_flat()
    ->intersect($f_0m4)
    ->MyFLAT::digits_increment(add => 1);
  my $case_3m4 = TurnLpred_flat()
    ->intersect($f_2m4)
    ->MyFLAT::digits_increment(add => 1);
  my $f = $case_0m4->union($case_1m4,
                           $case_2m4,
                           $case_3m4,
                           $f_one);
  $f = $f->MyFLAT::reverse;
  ok($f->equals(TurnRunSpred_LtoH_flat()));
  #  MyFLAT::FLAT_check_is_equal($f,TurnRunSpred_LtoH_flat());
}
{
  my $bad = 0;
  foreach my $n (0 .. 1024) {
    my $n2 = reverse cnv($n,10,2);
    my $by_flat = TurnRunSpred_LtoH_flat()->contains($n2) ? 1 : 0;
    my $by_func = TurnRunSpred($n);
    unless ($by_flat == $by_func) {
      print "n=$n [$n2] by_flat $by_flat by_func $by_func\n";
      last if $bad++ > 10;
    }
  }
  is($bad, 0, "TurnRunSpred_LtoH_flat() vs TurnRunSpred()");
}


#------------------------------------------------------------------------------
# dir4_flat()

ok (dir4_flat(0)->contains(''));
ok (dir4_flat(0)->contains('0'));
ok (dir4_flat(0)->contains('10'));

{
  my $bad = 0;
  foreach my $d (0 .. 3) {
    my $f = dir4_flat($d);
  N: foreach my $n (0 .. 256) {
      my $str = sprintf '%b', $n;
      $str =~ s/^0+//g;
      foreach my $zeros (0,1,3) {
        my $str = ('0'x$zeros) . $str;
        {
          my $by_flat = $f->contains($str) ? 1 : 0;
          my $by_dir  = (dir4($n) == $d);
          unless ($by_flat == $by_dir) {
            print "N=$n  $f->{name} on str=$str got=$by_flat want=$by_dir\n";
            last N if $bad++ > 10;
          }
        }
      }
    }
  }
  is($bad, 0, "dir4_flat() vs dir4()");
}


#------------------------------------------------------------------------------
# DpredFirst and DpredSecond

# first and second of double-visit are subsets of whole double-visits
ok(DpredFirstInf_flat()->is_subset_of(DpredInf_flat()));
ok(DpredSecond_flat()  ->is_subset_of(DpredInf_flat()));
{
  my $f = DpredFirstInf_flat()->union(DpredSecond_flat())->as_dfa;
  ok($f->equals(DpredInf_flat()),
     'DpredFirstInf_flat union DpredSecond_flat = DpredInf_flat');
  # MyFLAT::FLAT_check_is_equal($f, DpredInf_flat());
}


#------------------------------------------------------------------------------
# fBpred by hullsaround

# $hullsaround_segs_outer->print;
# $hullsaround_segs_inner->print;
# $hullsaround_segs_all->print;


#------------------------------------------------------------------------------
# hullsaround state counts not minimized

fBpred_hullsaround_init();
is($fRpred_hullsaround_flat->num_states, 116);
is($fLpred_hullsaround_flat->num_states, 116);
is($fBpred_hullsaround_flat->num_states, 116);

# no eventually accepting
is (scalar(@{$fBpred_state_enclosed_lists[0]}), 50);  # R
is (scalar(@{$fBpred_state_enclosed_lists[1]}), 17);  # L
is (scalar(@{$fBpred_state_enclosed_lists[2]}), 4);   # both B

is($fNonRpred_hullsaround_flat->MyFLAT::num_accepting, 78); # stated in alternate.tex
is($fNonLpred_hullsaround_flat->MyFLAT::num_accepting, 50); # incl eventually
is($fNonBpred_hullsaround_flat->MyFLAT::num_accepting, 32);

{
  my $f = MyFLAT::aref_to_FLAT_DFA
    (\@fBpred_state_table,
     name => 'fNonRpred_hullsaround',
     accepting_list => $fBpred_state_enclosed_lists[0],
    );
  is ($f->MyFLAT::num_accepting, 50);  # without eventually accepting
  my %info = $f->MyFLAT::get_eventually_accepting_info;
  is ($info{'depth'}, 2);
  is (scalar(@{$info{'states'}}), 28);
}
{
  my $f = MyFLAT::aref_to_FLAT_DFA
    (\@fBpred_state_table,
     name => 'fNonRpred_hullsaround',
     accepting_list => $fBpred_state_enclosed_lists[1],
    );
  is ($f->MyFLAT::num_accepting, 17);  # without eventually accepting
  my %info = $f->MyFLAT::get_eventually_accepting_info;
  is ($info{'depth'}, 2);
  is (scalar(@{$info{'states'}}), 33);
}
{
  my $f = MyFLAT::aref_to_FLAT_DFA
    (\@fBpred_state_table,
     name => 'fNonRpred_hullsaround',
     accepting_list => $fBpred_state_enclosed_lists[2],
    );
  is ($f->MyFLAT::num_accepting, 4);  # without eventually accepting
  my %info = $f->MyFLAT::get_eventually_accepting_info;
  is ($info{'depth'}, 2);
  is (scalar(@{$info{'states'}}), 28);
}


#------------------------------------------------------------------------------
# fRpred_flat() minimized equality

# is(fRpred_flat()->num_states, 30);
# is(fLpred_flat()->num_states, 32);
# 
# ok(fRpred_flat()->equals($fRpred_hullsaround_flat));
# ok(fLpred_flat()->equals($fLpred_hullsaround_flat));
# ok(fNonRpred_flat()->equals($fNonRpred_hullsaround_flat));
# ok(fNonLpred_flat()->equals($fNonLpred_hullsaround_flat));
# 
# ok( fNonRpred_flat()->equals($fNonRpred_hullsaround_flat->MyFLAT::minimize) );
# ok( fNonLpred_flat()->equals($fNonLpred_hullsaround_flat->MyFLAT::minimize) );


#------------------------------------------------------------------------------
# cases extras are on the boundary so are subset of fRpred, etc

# fBpred_hullsaround_init();
# foreach my $extra (fRpred_cases_extra_list()) {
#   ok($extra->is_subset_of($fRpred_hullsaround_flat));
# }
# foreach my $extra (fLpred_cases_extra_list()) {
#   ok($extra->is_subset_of($fLpred_hullsaround_flat));
# }
# foreach my $extra (fRpred_cases_extra_list(), fLpred_cases_extra_list()) {
#   ok($extra->is_subset_of($fBpred_hullsaround_flat));
# }

# # Extras subtract from NonBpred_k_flat().
# # Extra is bigger than NonBpred but has a non-empty intersection so that
# # it does subtract.
# foreach my $extra (fRpred_cases_extra_list(), fLpred_cases_extra_list()) {
#   my $int = NonBpred_k_flat()->intersect($extra);
#   ok(! $int->is_empty);
# }
# 
# # anything with Spred at one end of segment is fBpred,
# # including prefixes of Spred as they are not yet proven enclosed
# ok (Spred_k_flat()->is_subset_of($fBpred_hullsaround_flat));
# ok (Spred_k_flat()
#     ->MyFLAT::prefix
#     ->is_subset_of($fBpred_hullsaround_flat));


#------------------------------------------------------------------------------
# OXpred_flat()

# OXpred_flat()->MyFLAT::view;
{
  my $k = 8;
  my @a = grep {OXpred_flat()->contains(sprintf '%b',$_)} 0 .. 2**$k-1;
  is_deeply(\@a, [4, 8, 16, 20, 32, 40, 64, 80, 84, 128, 160, 168]);
}
{
  my $path = AlternatePaper_path();
  my $bad = 0;
 K: foreach my $k (0 .. 12) {
    my %xor;
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    foreach my $n ($n_lo .. $n_hi-1) {
      my ($x,$y) = $path->n_to_xy($n) or return;
      my @n_list = $path->xy_to_n_list($x,$y);
      @n_list = grep {$_<$n_hi} @n_list;
      @n_list==1 or @n_list==2 or die;
      if (@n_list == 2) {
        # an xor difference occurring
        push @{$xor{$n_list[0] ^ $n_list[1]}}, $n;
      }
    }
    foreach my $n ($n_lo .. $n_hi-1) {
      my $n2 = sprintf '%0*s', $k, cnv($n,10,2);
      my $by_flat = OXpred_flat()->contains($n2) ? 1 : 0;
      my $by_path = exists $xor{$n}  ? 1 : 0;
      if ($by_path != $by_flat) {
        print "k=$k xor n=$n2  by_flat=$by_flat by_path=$by_path",
          ($by_path ? ' (at ' . join(',',@{$xor{$n}}) . ')' : ()),
          "\n";
        last K if $bad++ > 10;
      }
    }
  }
  is($bad,0, "OXpred_flat() vs path");
}


#------------------------------------------------------------------------------
# RpredInf_flat()

ok(  Xpred_flat() ->contains('0'));
ok(  Xpred_flat() ->contains('1'));
ok(! Xpred_flat() ->contains('10'));
ok(! Xpred_flat() ->contains('010'));
ok(! Xpred_flat() ->contains('11'));
ok(  Xpred_flat() ->contains('100'));

ok(  RpredInf_flat() ->contains('0'));
ok(  RpredInf_flat() ->contains('1'));
ok(  RpredInf_flat() ->contains('10'));
ok(  RpredInf_flat() ->contains('11'));
ok(  RpredInf_flat() ->contains('100'));

ok(RpredInf_flat()->equals(NonRpredInf_flat()->complement));

{
  #              0 1 2 3 4 5 6              14
  my @samples = (1,1,1,1,1,1,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,0);
  foreach my $n (0 .. $#samples) {
    my $str = sprintf '%b', $n;
    foreach (0 .. 2) {  # leading zeros
      my $by_flat  = RpredInf_flat() ->contains($str) ? 1 : 0;
      is ($by_flat, $samples[$n],
          "RpredInf_flat() samples $n str $str");
      $str = "0$str";
    }
  }
}

#------------------------------------------------------------------------------
# TurnLpred_flat()
# TurnRpred_flat()

ok(! TurnLpred_flat() ->contains(''));
ok(! TurnRpred_flat()->contains(''));
ok(TurnLpred_flat() ->intersect(zero_digits_flat(2))->is_empty,
   "TurnLpred_flat() no match zero");
ok(TurnRpred_flat()->intersect(zero_digits_flat(2))->is_empty,
   "TurnLpred_flat() no match zero");
ok(TurnLpred_flat() ->intersect(TurnRpred_flat())->is_empty,
   "TurnLpred_flat() and TurnRpred_flat() disjoint");

{
  my $bad = 0;
  foreach my $n (1 .. 1000) {
    my $t_left  = TurnLpred($n);
    my $t_right = TurnRpred($n);
    my $str = sprintf '%b', $n;
    foreach (0 .. 2) {  # leading zeros
      my $f_left  = TurnLpred_flat() ->contains($str) ? 1 : 0;
      my $f_right = TurnRpred_flat()->contains($str) ? 1 : 0;
      if ($t_left != $f_left)  { $bad++; }
      if ($t_right != $t_right) { $bad++; }
      $str = "0$str";
    }
  }
  is($bad, 0, "TurnLpred_flat() and TurnRpred_flat() vs turn()");
}


#------------------------------------------------------------------------------
# dir4_LtoH_state_table

{
  my @dir4_LtoH_state_table = ([ 0,1 ],  # 0
                               [ 0,3 ],  # 1
                               [ 2,3 ],  # 2
                               [ 2,1 ],  # 3
                               [ 0,8 ],  # 4
                               [ 8,1 ],  # 5
                               [ 2,8 ],  # 6
                               [ 8,3 ],  # 7
                               [ 8,8]);  # 8
  my $bad = 0;
  foreach my $d (0 .. 3) {
    my $f = MyFLAT::aref_to_FLAT_DFA
      (\@dir4_LtoH_state_table,
       name => "dir=$d",
       accepting_list => [0,1,4],
       starting => 4+$d,
      );
  N: foreach my $n (0 .. 256) {
      my $str = sprintf '%b', $n;
      $str =~ s/^0+//g;
      foreach my $zeros (0,1,3) {
        my $str = reverse( ('0'x$zeros) . $str );

        {
          my $by_flat = $f->contains($str) ? 1 : 0;
          my $by_dir = (dir4($n) == $d);
          unless ($by_flat == $by_dir) {
            print "N=$n  $f->{name} on str=$str got=$by_flat want=$by_dir\n";
            last N if $bad++ > 10;
          }
        }
      }
    }
  }
  is($bad, 0, "dir4pred vs dir4()");
}


#------------------------------------------------------------------------------
# TAparentDir

{
  my @TAparentDir_samples = (0, 2, 3, 1, 0, 3, 0, 2, 1, 2, 3, 3, 0, 2, 3, 1, 0);
  foreach my $i (0 .. $#TAparentDir_samples) {
    my $n = $i + 1;
    my $sample = $TAparentDir_samples[$i];
    is (TAparentDir($n), $sample, "TAparentDir");

    foreach my $dir (0 .. 2) {
      my $HtoL = TAparentDir_flat($dir);
      my $LtoH = TAparentDir_LtoH_flat($dir);
      foreach my $zeros (0 .. 0) {
        my $str = sprintf '%b', $n;
        $str = ('0'x$zeros) . $str;
        ok((!!$HtoL->contains($str)) == ($sample==$dir),
           "HtoL dir=$dir str=$str");
        $str = reverse $str;
        # ok((!!$LtoH->contains($str)) == ($sample==$dir),
        #    "LtoH dir=$dir str=$str");
      }
    }
  }
}

{
  my $re = FLAT::Regex->new ('((0|1)* 10 | 0 | []) (00|11)* 1')->as_dfa;
  ok(TAparentDir_flat(0)->equals($re));
}
{
  my $re = FLAT::Regex->new ('((0|1)* 0 | []) 1 (00|11)* 0  0 1*')->as_dfa;
  ok(TAparentDir_flat(1)->equals($re));
}
{
  my $re = FLAT::Regex->new ('((0|1)* 0 | []) 1 (00|11)* 0')->as_dfa;
  ok(TAparentDir_flat(2)->equals($re));
}

foreach my $dir (0 .. 3) {
  my $HtoL = TAparentDir_flat($dir);
  my $LtoH = TAparentDir_LtoH_flat($dir);

  # N=0 has no parent, but dir=3 HtoL does match N=0
  my $zeros_flat = MyFLAT::zero_digits_flat(2);
  $HtoL = $HtoL->difference($zeros_flat)->as_dfa;

  ok($LtoH->equals($HtoL->MyFLAT::reverse), "dir=$dir");

  # FLAT_show_breadth($LtoH,4);
  # FLAT_show_breadth($HtoL,4);
  # FLAT_check_is_equal($LtoH, $HtoL->MyFLAT::reverse);
}


#------------------------------------------------------------------------------
# fXpred, fGpred, fVpred

{
  # fXpred

  my $f_one       = MyFLAT::fraction_digits(1,1, radix => 2);
  my $f_two_third = MyFLAT::fraction_digits(2,3, radix => 2);
  my $f_digits_0_1 = FLAT::Regex->new('(00|01)*')->as_dfa;
  my $f_digits_0_1_odd = FLAT::Regex->new('(00|01)* 01')->as_dfa;
  my $f_odd_plus_23 = $f_digits_0_1_odd->MyFLAT::concat($f_two_third);

  my $f = $f_one->union($f_digits_0_1->MyFLAT::fraction_also_nines(radix=>2),
                        $f_odd_plus_23)
    ->as_dfa;
  $f = $f->MyFLAT::prefix;
  ok($f->equals(fXpred_flat()));

  my $fNon = $f->complement;
  ok($fNon->equals(fNonXpred_flat()));

  # f=1/3 is covered by case base-4 digits 0,1
  my $f_one_third = MyFLAT::fraction_digits(1,3, radix => 2);
  ok ($f_one_third->is_subset_of($f_digits_0_1->MyFLAT::prefix));
}
{
  # fGpred

  my $f_digits_0_2 = FLAT::Regex->new('(00|10)*')->as_dfa;
  my $f_frac_56 = MyFLAT::fraction_digits(5,6, radix => 2);
  my $f_digits_plus_56 = $f_digits_0_2->MyFLAT::concat($f_frac_56);

  my $f = $f_digits_0_2
    ->MyFLAT::prefix
    ->MyFLAT::fraction_also_nines(radix=>2)
    ->union($f_digits_plus_56)
    ->as_dfa;
  $f = $f->MyFLAT::prefix;
  ok($f->equals(fGpred_flat()));

  my $fNon = $f->complement;
  ok($fNon->equals(fNonGpred_flat()));
}
{
  # fVpred

  # 101111... = 110000...

  my $f_one       = MyFLAT::fraction_digits(1,1, radix => 2);
  my $f_one_third = MyFLAT::fraction_digits(1,3, radix => 2);
  my $f_two_third = MyFLAT::fraction_digits(2,3, radix => 2);
  my $f_digits_2_3      = FLAT::Regex->new('(10|11)*')->as_dfa;
  my $f_digits_2_3_zeros = FLAT::Regex->new('(10|11)* 11 0*')->as_dfa;
  my $f_digits_2_3_even  = FLAT::Regex->new('(10|11)* 10')->as_dfa;
  my $f_digits_plus_13 = $f_digits_2_3_even->MyFLAT::concat($f_one_third);

  # my $f_digits_2_3_sub_1 = FLAT::Regex->new('(10|11)* (01|10)')->as_dfa;
  # my $f_digits_sub_13 = $f_digits_2_3_sub_1->MyFLAT::concat($f_two_third);

  my $f = $f_one
    ->union($f_one_third,
            $f_digits_2_3,
            $f_digits_2_3_zeros,
            $f_digits_plus_13)
    ->as_dfa;
  $f = $f->MyFLAT::prefix;
  ok($f->equals(fVpred_flat()));

  my $fNon = $f->complement;
  ok($fNon->equals(fNonVpred_flat()));

  # $fNon->MyFLAT::minimize->MyFLAT::view;
  # fNonVpred_flat()->MyFLAT::minimize->MyFLAT::view;
  # MyFLAT::FLAT_check_is_equal($f,fVpred_flat()->MyFLAT::prefix);
}

#------------------------------------------------------------------------------
# R,L,B squares

# ok_conf_expand_within_self($R_square_segs);
# ok_conf_expand_within_self($L_square_segs);
ok_conf_expand_within_self($B_squares_segs);


#------------------------------------------------------------------------------
# Epred runs

# Last segment before unfold point 2^k is not an enclosure, so not part of a
# run, so no worry that Epred would be 10000 which is not within k bits.

ok(DpredSecond_flat()->equals(Epred_flat()->MyFLAT::digits_increment));

{
  my $f = Epred_flat();
  for (1 .. 7) {
    $f = Epred_flat()->intersect($f->MyFLAT::digits_increment)
      ->MyFLAT::minimize(verbose=>1);
    ok(! $f->is_empty);
  }
  # MyFLAT::FLAT_show_breadth($f, 8);
  my $max_k = 20;
  is(join(',', MyFLAT::FLAT_count_contains($f,$max_k)),
     '0,0,0,0,0,0,0,1,1,5,7,21,35,85,155,341,651,1365,2667,5461,10795');

  # one further for n+8 is 9 terms n to n+8 inclusive
  $f = Epred_flat()->intersect($f->MyFLAT::digits_increment);
  ok($f->is_empty);
}


#------------------------------------------------------------------------------
# Turn3left Turn3right counts consecutive turns

{
  my $f = TurnLpred_flat();
  for (1 .. 2) {
    # MyFLAT::FLAT_show_breadth($f, 4);
    # MyFLAT::FLAT_show_breadth($f->MyFLAT::digits_increment, 4);
    $f = TurnLpred_flat()->intersect($f->MyFLAT::digits_increment);
  }
  # MyFLAT::FLAT_show_breadth($f, 6);
  my $max_k = 10;
  is(join(',', FLAT_count_contains($f,$max_k)),
     join(',', map {Turn3left($_)} 0 .. $max_k));
}
{
  my $f = TurnRpred_flat();
  for (1 .. 2) {
    $f = TurnRpred_flat()->intersect($f->MyFLAT::digits_increment);
  }
  my $max_k = 10;
  is(join(',', FLAT_count_contains($f,$max_k)),
     join(',', map {Turn3right($_)} 0 .. $max_k));
}


#------------------------------------------------------------------------------
# Xpred_flat, Gpred_flat

ok(Xpred_flat()->contains(''));
ok(Gpred_flat()->contains(''));
ok(SpredInf_flat()->contains(''));

ok(Xpred_flat()->contains('0'));
ok(Gpred_flat()->contains('0'));
ok(SpredInf_flat()->contains('0'));


{
  my $max_k = 10;
  ### count: FLAT_count_contains(DpredSecond_flat(),$max_k)
  ### D    : map {D($_)} 0 .. $max_k

  # D counts locations x,y and there are 2 N's there so Dpred = 2*D
  is_deeply([ FLAT_count_contains(DpredSecond_flat(),$max_k) ],
            [ map {D($_)} 0 .. $max_k ]);
}
{
  my $path = Math::PlanePath::AlternatePaper->new;
  my $bad = 0;
  my $n_hi = 256;
 N: foreach my $n (0 .. $n_hi) {
    my $str = sprintf '%b', $n;
    foreach my $zeros (0 .. 3) {
      my $str = ('0' x $zeros) . $str;
      my ($x,$y) = $path->n_to_xy($n);
      my @n_list = $path->xy_to_n_list($x,$y);
      {
        my $f = Xpred_flat();
        my $f_pred = $f->contains($str) ? 1 : 0;
        my $p_pred = ($y==0 ? 1 : 0);
        if ($p_pred != $f_pred) {
          print "$f->{name} at N=$n $str got flat=$f_pred want path=$p_pred\n";
          last N if $bad++ > 10;
        }
      }
      {
        my $f = Gpred_flat();
        my $f_pred = $f->contains($str) ? 1 : 0;
        my $p_pred = ($x==$y ? 1 : 0);
        if ($p_pred != $f_pred) {
          print "$f->{name} at N=$n $str got flat=$f_pred want path=$p_pred\n";
          last N if $bad++ > 10;
        }
      }
      {
        my $f = SpredInf_flat();
        my $f_pred = $f->contains($str) ? 1 : 0;
        my $p_pred = @n_list == 1 ? 1 : 0;
        if ($p_pred != $f_pred) {
          print "$f->{name} at N=$n $str got flat=$f_pred want path=$p_pred\n";
          last N if $bad++ > 10;
        }
      }
      {
        my $f = DpredInf_flat();
        my $f_pred = $f->contains($str) ? 1 : 0;
        my $p_pred = @n_list == 2 ? 1 : 0;
        if ($p_pred != $f_pred) {
          print "$f->{name} at N=$n $str got flat=$f_pred want path=$p_pred\n";
          last N if $bad++ > 10;
        }
      }
      {
        my $f = DpredSecond_flat();
        my $f_pred = $f->contains($str) ? 1 : 0;
        my $p_pred = @n_list == 2 && $n==$n_list[1] ? 1 : 0;
        if ($p_pred != $f_pred) {
          print "$f->{name} at N=$n $str got flat=$f_pred want path=$p_pred\n";
          last N if $bad++ > 10;
        }
      }
    }
  }
  is($bad, 0, "Xpred etc vs path");
}


# #------------------------------------------------------------------------------
# 
# my $NonRpredInf_eventually_flat = MyFLAT::eventually_accepting(NonRpredInf_flat());
# # my $NonLpred_k_eventually_flat =MyFLAT::eventually_accepting(NonLpred_k_flat());
# # my $NonBpred_k_eventually_flat =MyFLAT::eventually_accepting(NonBpred_k_flat());

# #------------------------------------------------------------------------------
# # fNonRpred etc subsets of NonRpred eventually
# 
# # fNonRpred has its right side square enclosed, so subset of NonRpred,
# # or rather of NonRpred eventually enclosed
# # ok (fNonRpred_flat()->is_subset_of($NonRpredInf_eventually_flat));
# # ok (fNonLpred_flat()->is_subset_of($NonLpred_k_eventually_flat));
# # ok (fNonBpred_flat()->is_subset_of($NonBpred_k_eventually_flat));


#------------------------------------------------------------------------------
# Spred_k_flat(), Dpred_k_flat()

# {
#   my $max_k = 10;
#   ### list: FLAT_count_contains(Dpred_k_flat(),$max_k)
#   ### D: map {D($_)} 0 .. $max_k
# 
#   # D counts locations x,y and there are 2 N's there so Dpred = 2*D
#   is_deeply([ FLAT_count_contains(Dpred_k_flat(),$max_k) ],
#             [ map {2*D($_)} 0 .. $max_k ]);
# 
#   # # Spred_k_flat doesn't include N=2^k, so S-1
#   # is_deeply([ FLAT_count_contains(Spred_k_flat(),$max_k) ],
#   #           [ map {S($_)-1} 0 .. $max_k ]);
# }
# 
# ok(SpredLeft_runs_flat()->contains(''));

# # anything with Spred at one end of segment is boundary
# # (but only subset as boundary has some with doubles at both ends)
# ok(Spred_k_flat()->is_subset_of( Bpred_k_flat() ));
# 
# ok(Spred_k_flat()->equals( Dpred_k_flat()->complement ));
# ok(Spred_k_flat()->intersect(Dpred_k_flat())->is_empty); # disjoint


#------------------------------------------------------------------------------
# DpredLeft and DpredRight

# ok(DpredRight_k_flat()->is_subset_of(TurnRpred_flat()));
# ok(DpredLeft_k_flat() ->is_subset_of(TurnLpred_flat()));
# ok(SpredRight_k_flat()->is_subset_of(TurnRpred_flat()));
# ok(SpredLeft_k_flat() ->is_subset_of(TurnLpred_flat()));
# 
# # left, right disjoint
# ok(SpredLeft_k_flat() ->intersect(SpredRight_k_flat())->is_empty);
# ok(DpredLeft_k_flat() ->intersect(DpredRight_k_flat())->is_empty);
# 
# # whole = left + right
# ok(Dpred_k_flat()->equals( DpredRight_k_flat()
#                            ->union(DpredLeft_k_flat()) ->as_dfa ));
# ok(Spred_k_flat()->equals( SpredRight_k_flat()
#                            ->union(SpredLeft_k_flat())
#                            ->union(zero_digits_flat(2))   ->as_dfa ));
# 
# ok(  Spred_k_flat()->contains('111'));  # N=7
# ok(! Dpred_k_flat()->contains('111'));
# ok(  Dpred_k_flat()->contains(  '0111'));  # in k=4
# ok(  Dpred_k_flat()->contains( '00111'));  # in k=5
# ok(  Dpred_k_flat()->contains('000111'));  # in k=6
# 
# # double-visiteds in k are a subset of those in k+1,
# # with extra doubles at the join
# foreach my $f (Dpred_k_flat(),
#                DpredLeft_k_flat(),
#                DpredRight_k_flat()) {
#   my $c = FLAT::Regex->new('0')->as_dfa->MyFLAT::concat($f)->as_min_dfa;
#   ok($c->is_subset_of($f),
#      "$f->{name} extra leading zero, k is subset of k+1");
# }
# 
# foreach my $len (0 .. 3) {          # N=0
#   my $str = '0' x $len;
#   ok(Spred_k_flat()     ->contains($str));
#   ok(! SpredRight_k_flat()->contains($str));
#   ok(! SpredLeft_k_flat() ->contains($str));
#   ok(! Dpred_k_flat()     ->contains($str));
#   ok(! DpredLeft_k_flat() ->contains($str));
#   ok(! DpredRight_k_flat()->contains($str));
# }
# 
# {
#   my $path = Math::PlanePath::AlternatePaper->new;
#   my $bad = 0;
#  K: foreach my $k (0 .. 10) {
#     my $n_hi = 1 << $k;
#     ### $n_hi
#     foreach my $n (0 .. $n_hi) {
#       my $str = sprintf '%0*b', $k, $n;
#       my ($x,$y) = $path->n_to_xy($n);
#       my @n_list = $path->xy_to_n_list($x,$y);
#       my $turn = $n>=1 && turn($n);
#       my $t_left  = $turn==1;
#       my $t_right = $turn==-1;
#       ### @n_list
#       ### $turn
# 
#       # no DpredInf_flat() yet
#       # {
#       #   my $p_pred = @n_list == 2;
#       #   my $f_pred = DpredInf_flat()->contains($str) ? 1 : 0;
#       #   if ($p_pred != $f_pred) {
#       #     print "Dpred_k_flat() at N=$n $str got $f_pred want path=$p_pred\n";
#       #     last K if $bad++ > 10;
#       #   }
#       # }
# 
#       @n_list = grep {$_ <= $n_hi} @n_list;
#       {
#         my $f = Dpred_k_flat();
#         my $p_pred = @n_list == 2 ? 1 : 0;
#         my $f_pred = $f->contains($str) ? 1 : 0;
#         if ($p_pred != $f_pred) {
#           print "$f->{name} at N=$n $str got $f_pred want path=$p_pred\n";
#           last K if $bad++ > 10;
#         }
#       }
#       {
#         my $f = DpredLeft_k_flat();
#         my $p_pred = @n_list == 2 && $t_left ? 1 : 0;
#         my $f_pred = $f->contains($str) ? 1 : 0;
#         if ($p_pred != $f_pred) {
#           print "$f->{name} at N=$n $str got $f_pred want path=$p_pred\n";
#           last K if $bad++ > 10;
#         }
#       }
#       {
#         my $f = DpredRight_k_flat();
#         my $p_pred = @n_list == 2 && $t_right ? 1 : 0;
#         my $f_pred = $f->contains($str) ? 1 : 0;
#         if ($p_pred != $f_pred) {
#           print "$f->{name} at N=$n $str got $f_pred want path=$p_pred\n";
#           last K if $bad++ > 10;
#         }
#       }
#     }
#   }
#   is($bad, 0, "Dpred_flat() vs path");
# }
# # for(n=0,31,if(Dpred(n),print(to_binary(n))))

# sub check_flat_vs_predicate {
#   my ($flat, $func, $name) = @_;
#   my $bad = 0;
#   foreach my $k (1 .. 10) {
#     foreach my $n (0 .. 2**$k-1) {
#       my $str = sprintf '%0*b', $k, $n;
#       my $got = $flat->contains($str) ? 1 : 0;
#       my $want = $func->($k,$n) ? 1 : 0;
#       if ($got != $want) {
#         print "$name k=$k n=$n $str got $got want $want\n";
#         exit if $bad++ > 10;
#       }
#     }
#   }
# }
#   check_flat_vs_predicate($DpredLeft_k_flat, \&DpredLeft_by_path, 'DpredLeft');
#   check_flat_vs_predicate($DpredRight_k_flat, \&DpredRight_by_path, 'DpredRight');
#   check_flat_vs_predicate($SpredLeft_flat, \&SpredLeft_by_path, 'SpredLeft');
#   check_flat_vs_predicate($SpredRight_flat, \&SpredRight_by_path, 'SpredRight');

# sub Spred_k_by_path {
#   my ($k,$n) = @_;
#   my ($x,$y) = $path->n_to_xy($n);
#   my @n_list = $path->xy_to_n_list($x,$y);
#   my ($n_lo, $n_hi) = $path->level_to_n_range($k);
#   @n_list = grep {$_ <= $n_hi} @n_list;
#   return @n_list==1;
# }
# sub SpredLeft_by_path {
#   my ($k,$n) = @_;
#   return $n>=1 && turn($n)==1 && Spred_k_by_path($k,$n);
# }
# sub SpredRight_by_path {
#   my ($k,$n) = @_;
#   return $n>=1 && turn($n)==-1 && Spred_k_by_path($k,$n);
# }
# sub DpredRight_by_path {
#   my ($k,$n) = @_;
#   return $n>=1 && turn($n)==-1 && ! Spred_k_by_path($k,$n);
# }
# sub DpredLeft_by_path {
#   my ($k,$n) = @_;
#   return $n>=1 && turn($n)==1 && ! Spred_k_by_path($k,$n);
# }


#------------------------------------------------------------------------------
# Bpred_k_flat()

# ok(Bpred_k_flat()->equals( NonBpred_k_flat()->complement ));
# 
# is(ref Bpred_k_flat(), 'FLAT::DFA');
# {
#   my @starting = Bpred_k_flat()->get_starting;
#   is(scalar(@starting), 1);
# }


#------------------------------------------------------------------------------
# Lpred_k_flat()

# {
#   my $max_k = 10;
#   is_deeply([ FLAT_count_contains(Lpred_k_LtoH_flat(),$max_k) ],
#             [ map {L($_)} 0 .. $max_k ]);
# 
#   is_deeply([ FLAT_count_contains(Lpred_k_flat(),$max_k) ],
#             [ map {L($_)} 0 .. $max_k ]);
# }
# 
# ok(Lpred_k_flat()->equals( NonLpred_k_flat()->complement ));
# ok(NonLpred_k_LtoH_flat()->MyFLAT::reverse->equals(NonLpred_k_flat()));
# ok(Lpred_k_LtoH_flat()   ->MyFLAT::reverse->equals(Lpred_k_flat()));
# 
# {
#   my $path = Math::PlanePath::AlternatePaper->new;
#   my $bad = 0;
#  K: foreach my $k (0 .. 10) {
#     my $n_hi = 1 << $k;
#     foreach my $n (0 .. 2**$k-1) {
#       my $str = sprintf '%0*b', $k, $n;
#       my $p_pred = path_n_is_left_boundary($path,$n,$n_hi) ?1:0;
# 
#       {
#         my $f_pred = Lpred_k_flat()->contains($str) ? 1 : 0;
#         if ($p_pred != $f_pred) {
#           print "Lpred_k_flat() at N=$n $str got $f_pred want path=$p_pred\n";
#           last K if $bad++ > 10;
#         }
#       }
#       {
#         my $LtoH_str = reverse $str;
#         my $f_LtoH_pred = Lpred_k_LtoH_flat()->contains($LtoH_str) ? 1 : 0;
#         if ($p_pred != $f_LtoH_pred) {
#           print "Lpred_k_LtoH_flat() at N=$n $LtoH_str got $f_LtoH_pred want path=$p_pred\n";
#           last K if $bad++ > 10;
#         }
#       }
#     }
#   }
#   is($bad, 0, "Lpred_flat() vs path");
# }
# # for(n=0,31,if(Lpred(n),print(to_binary(n))))
# 
# ok(Lpred_re_flat()   ->equals(Lpred_k_flat()));
# ok(NonLpred_re_flat()->equals(NonLpred_k_flat()));


#------------------------------------------------------------------------------


sub path_n_is_right_boundary {
  my ($path, $n, $n_limit) = @_;
  ### path_n_is_right_boundary(): $n
  ### $n_limit
  if (defined $n_limit && $n >= $n_limit) {
    croak "path_n_is_right_boundary() $n >= limit $n_limit";
  }
  my ($x1,$y1) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  ($dx,$dy) = xy_rotate_minus90($dx,$dy);  # to right side
  my $x1s = $x1 + $dx;
  my $y1s = $y1 + $dy;
  my $x2s = $x2 + $dx;
  my $y2s = $y2 + $dy;
  return (! path_xyxy_is_traversed_either($path, $x1,$y1, $x1s,$y1s, $n_limit)
          || ! path_xyxy_is_traversed_either($path, $x2,$y2, $x2s,$y2s, $n_limit)
          || ! path_xyxy_is_traversed_either($path, $x1s,$y1s, $x2s,$y2s, $n_limit));
}
sub path_n_is_left_boundary {
  my ($path, $n, $n_limit) = @_;
  ### path_n_is_left_boundary(): $n
  if ($n >= $n_limit) {
    croak "path_n_is_left_boundary() $n >= limit $n_limit";
  }
  my ($x1,$y1) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  ($dx,$dy) = xy_rotate_plus90($dx,$dy);  # to left side
  my $x1s = $x1 + $dx;
  my $y1s = $y1 + $dy;
  my $x2s = $x2 + $dx;
  my $y2s = $y2 + $dy;
  if (! path_xyxy_is_traversed_either($path, $x1,$y1, $x1s,$y1s, $n_limit)) {
    ### x1 y1 to s ...
    return 1;
  }
  if (! path_xyxy_is_traversed_either($path, $x2,$y2, $x2s,$y2s, $n_limit)) {
    ### x2 y2 to s ...
    return 1;
  }
  if (! path_xyxy_is_traversed_either($path, $x1s,$y1s, $x2s,$y2s, $n_limit)) {
    ### s to s ...
    return 1;
  }
  ### not left boundary ...
  return 0;
}
sub path_xyxy_is_traversed_either {
  my ($path, $x1,$y1, $x2,$y2, $n_limit) = @_;
  ### xyxy_is_traversed_either(): "$x1,$y1 to $x2,$y2"
  my $arms = $path->arms_count;
  foreach my $n1 ($path->xy_to_n_list($x1,$y1)) {
    next if defined $n_limit && $n1 > $n_limit;
    foreach my $n2 ($path->xy_to_n_list($x2,$y2)) {
      next if defined $n_limit && $n2 > $n_limit;
      if (abs($n1-$n2) == $arms) {
        ### yes: "$n1 to $n2"
        return 1;
      }
    }
  }
  return 0;
}


#------------------------------------------------------------------------------
exit 0;
