#!/usr/bin/perl -w

# Copyright 2013, 2014, 2015, 2016, 2017, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.


# Usage: perl test.pl
#
# Run some self-tests on ../alternate.tex and on MyAlternate.pm.
#

use 5.010;
use strict;
use warnings;
use File::Slurp;
use IPC::Run;
use List::Util 'sum';

use FindBin;
use lib $FindBin::Bin;
use MyAlternate ':all';


my $source_dir = File::Spec->catdir ($FindBin::Bin, File::Spec->updir);
$source_dir = File::Spec->rel2abs($source_dir);
print "source dir $source_dir\n";

chdir $source_dir or die $!;

my $alternate_tex_filename = 'alternate.tex';
# $alternate_tex_filename = 'devel/more.tex';
my $alternate_pdf_filename = 'alternate.pdf';
my $alternate_tex_str = File::Slurp::slurp($alternate_tex_filename);
print "tex ",count_lines($alternate_tex_str)," lines\n";

my $alternate_tex_str_sans_comments = $alternate_tex_str;
$alternate_tex_str_sans_comments =~ s/((^|[^\\])(\\\\)*)%[^\n]*/$1/mg;
print count_lines($alternate_tex_str_sans_comments)," lines\n";


#------------------------------------------------------------------------------
# OEIS refs

{
  my %anums;
  while ($alternate_tex_str_sans_comments
         =~ /My[a-zA-Z]*Anumber.*?(?<anum>A\d{6,})/sg) {
    $anums{$+{'anum'}}++;
  }
  my $total = sum(0, values %anums);
  my $distinct = keys %anums;

  my $idx_str = File::Slurp::read_file
    (File::Spec->catfile($source_dir,'dragon.anums.idx'),
     err_mode=>'quiet') // '';
  my $idx_num_lines = str_num_lines($idx_str);

  my %ind_anums;
  my $ind_str = File::Slurp::read_file
    (File::Spec->catfile($source_dir,'dragon.anums.ind'),
     err_mode=>'quiet') // '';
  while ($ind_str =~ /\\item.*?(A\d{6,})\b/sg) { $ind_anums{$1}++; }
  my $ind_num_items = keys %ind_anums;
  $ind_num_items == sum(0, values %ind_anums)
    or die "oops, duplicate in dragon.anums.ind";
  foreach my $anum (keys %ind_anums) {
    if (! exists $anums{$anum}) {
      print "ind extra over tags: $anum\n";
      # $bad++;
    }
  }

  print "A-number tags $total, distinct $distinct, cf idx $idx_num_lines ind $ind_num_items\n";
}

sub str_num_lines {
  my ($str) = @_;
  return ($str =~ tr/\n/\n/);
}


#------------------------------------------------------------------------------
# ../devel/alternate-defines.gp
# ../devel/alternate-defines.header.gp

{
  # same copyright years as alternate.tex where defines are extracted from
  $alternate_tex_str =~ /^% (Copyright .*)/m or die;
  my $want_copyright = $1;
  foreach my $filename ('devel/alternate-defines.gp',
                        'devel/alternate-defines.header.gp') {
    my $str = File::Slurp::slurp($filename);
    $str =~ /^\\\\ \Q$want_copyright\E$/m
      or die "$filename copyright different from alternate.tex";
  }
}

#------------------------------------------------------------------------------
# alternate.pdf Title and Author

{
  IPC::Run::run(['pdfinfo', '-isodates', $alternate_pdf_filename],
                '>', \my $pdfinfo_out)
      or die;
  $pdfinfo_out =~ s/^Jav.*\n//m;
  print $pdfinfo_out;
  {
    $pdfinfo_out =~ /^Title:\s*(.*)/m
      or die "pdf no Title";
    my $title = $1;
    # print "title: $title\n";
    $title =~ /Alternate/
      or die;
  }
  {
    $pdfinfo_out =~ /^Author:\s*(.*)/m
      or die "pdf no Author";
    my $author = $1;
    # print "author: $author\n";
    $author =~ /Kevin/
      or die;
  }
}

#------------------------------------------------------------------------------
sub pos_linenum {
  my ($str) = @_;
  $str = substr($str, 0, pos($_[0]));
  return 1 + scalar($str =~ tr/\n//);
}

sub count_lines {
  my ($str) = @_;
  my $count = ($str =~ tr/\n/\n/);
  if ($str =~ /[^\n]$/) { $count++; }
  return $count;
}

#------------------------------------------------------------------------------
exit 0;
